#include <thread>
#include "headers.hpp"

#ifdef CURL_SUPPORT
#include <curl/curl.h>

// Used to hide cURL output
size_t writeNothing(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
   return nmemb;
}

void sendUsageData(tuple<string, string> t)
{
    string language = get<0>(t);
    string compiler = get<1>(t);

    CURL *curl = curl_easy_init();
    if (curl)
    {
        curl_httppost *formpost = NULL;
        curl_httppost *lastptr = NULL;
        curl_slist *headerlist = NULL;

        char strV[10]; sprintf(strV, "%f", VERSION);

        curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "version",
                     CURLFORM_COPYCONTENTS, strV, CURLFORM_END);

        curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "language",
                     CURLFORM_COPYCONTENTS, language.c_str(), CURLFORM_END);

        curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "compiler",
                     CURLFORM_COPYCONTENTS, compiler.c_str(), CURLFORM_END);

        Output::debug("Sending usage statistics...");

        curl_easy_setopt(curl, CURLOPT_URL, "tmake.tim32.org/submit");
        //curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, &writeNothing);

        headerlist = curl_slist_append(headerlist, "Expect:");
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headerlist);

        curl_easy_setopt(curl, CURLOPT_HTTPPOST, formpost);

        CURLcode res = curl_easy_perform(curl);

        if (res != 0)
        {
            string err = curl_easy_strerror(res);
            Output::warning("cURL error while sending stats: " + err);
        }

        curl_formfree(formpost);
        curl_slist_free_all(headerlist);
        curl_easy_cleanup(curl);
    }

    //pthread_exit(NULL);
}

#endif

Project::Project(lua_State *L)
{
    luaState = L;

    generatedFiles.push_back(Arch::buildDir());
    generatedFiles.push_back(".tmake.cache/");

    compiler = NULL;
}

Project::~Project()
{
}

void Project::reset()
{
    if (compiler)
    {
        compiler->reset();
    }

    compiler = NULL;

    sources.clear();
    headers.clear();
    resources.clear();
    libraries.clear();
    includeDirs.clear();
    libraryDirs.clear();
    definitions.clear();
    cflags.clear();
    lflags.clear();

    includeDirs.push_back("include/");

    addDefinition("ARCH=" + Arch::name());
    addDefinition(Str::upper(OS::name()));
}

string Project::getCompiledName(string type, string executable)
{
    string n = compiler->getComputerName(executable, type);
    if (n.empty())
    {
        n = OS::getName(type, executable);
    }

    return n;
}

void Project::setName(string n)
{
    name = n;

    addDefinition("BINDIR=" + binDir(name));
    addDefinition("LIBDIR=" + libDir(name));
    addDefinition("INCDIR=" + incDir(name));
    addDefinition("CONFDIR=" + confDir(name));
    addDefinition("DATADIR=" + dataDir(name));
}

string Project::getName()
{
    return name;
}

void Project::setLanguage(string l)
{
    language = l;
    compiler = tmake->chooseCompiler(language);

#ifdef CURL_SUPPORT
    tuple<string, string> t(language, compiler->getName());
#ifdef PTHREADS_SUPPPORT
    pthread_t thread;
    int ret = pthread_create(&thread, NULL, sendUsageData, t);
    if (ret != 0)
    {
        Output::warning("Multi-threading failed...");
        sendUsageData(t);
    }
#else
    sendUsageData(t);
#endif
#endif
}

string Project::getLanguage()
{
    return language;
}

void Project::setCompiler(string n)
{
    compiler = tmake->getCompiler(n, language);
}

void Project::addSource(string filename, bool generated)
{
    Output::debug("Adding source file: " + filename);

    string fullFilename = OS::convertPath(tmake->srcDir + filename);
    if (FS::fileExists(fullFilename))
    {
        sources.push_back(OS::convertPath(filename));

        if (generated)
        {
            generatedFiles.push_back(fullFilename);
        }
    }
    else
    {
        Lua::error(luaState, "Source file, " + fullFilename + ", doesn't exist.");
    }
}

void Project::addHeader(string filename, bool generated)
{
    Output::debug("Adding header file: " + filename);

    string fullFilename = OS::convertPath(tmake->headerDir + filename);
    if (FS::fileExists(fullFilename))
    {
        headers.push_back(OS::convertPath(filename));

        if (generated)
        {
            generatedFiles.push_back(fullFilename);
        }
    }
    else
    {
        Lua::error(luaState, "Header file, " + fullFilename + ", doesn't exist.");
    }
}

void Project::addResource(string filename, bool generated)
{
    Output::debug("Adding resource file: " + filename);

    string fullFilename = OS::convertPath(tmake->resDir + filename);
    if (FS::fileExists(fullFilename))
    {
        resources.push_back(OS::convertPath(filename));

        if (generated)
        {
            generatedFiles.push_back(fullFilename);
        }
    }
    else
    {
        Lua::error(luaState, "Resource file, " + fullFilename + ", doesn't exist.");
    }
}

void Project::addPackage(string name, float minVersion, float maxVersion)
{
    Package *package = tmake->getPackage(name);
    if (!packageExists(name))
    {
        Lua::error(luaState, "Failed to find package, " + package->getErrorMsg());
    }

    if (!package->correctVersion(minVersion, maxVersion))
    {
        Lua::error(luaState, "Package doesn't match version requirements.");
    }

    for (unsigned int i = 0; i < package->libraryDirs.size(); i++)
    {
        addLibraryDir(package->libraryDirs[i]);
    }

    for (unsigned int i = 0; i < package->libraries.size(); i++)
    {
        addLibrary(package->libraries[i]);
    }

    for (unsigned int i = 0; i < package->includeDirs.size(); i++)
    {
        addIncludeDir(package->includeDirs[i]);
    }

    for (unsigned int i = 0; i < package->definitions.size(); i++)
    {
        addDefinition(package->definitions[i]);
    }

    for (unsigned int i = 0; i < package->cflags.size(); i++)
    {
        addCFlag(package->cflags[i]);
    }

    for (unsigned int i = 0; i < package->lflags.size(); i++)
    {
        addLFlag(package->lflags[i]);
    }
}

void Project::addLibrary(string name)
{
    if (!name.empty())
    {
        if (!libraryExists(name))
        {
            Lua::error(luaState, "Failed to find library, " + name);
        }

        libraries.push_back(name);
    }
}

void Project::addIncludeDir(string name)
{
    if (!name.empty())
    {
        if (!includeDirExists(name))
        {
            Lua::error(luaState, "Failed to find include dir, " + name);
        }

        includeDirs.push_back(name);
    }
}

void Project::addLibraryDir(string name)
{
    if (!name.empty())
    {
        if (!libraryDirExists(name))
        {
            Lua::error(luaState, "Failed to find library dir, " + name);
        }

        libraryDirs.push_back(name);
    }
}

void Project::addDefinition(string name)
{
    if (!name.empty())
    {
        definitions.push_back(name);
    }
}

void Project::addCFlag(string name)
{
    if (!name.empty())
    {
        cflags.push_back(name);
    }
}

void Project::addLFlag(string name)
{
    if (!name.empty())
    {
        lflags.push_back(name);
    }
}

void Project::addOption(string name, string description, bool defaultValue)
{
    if (!tmake->cache->optionExists(name))
    {
        bool b = false;
        string s;
        if (defaultValue)
        {
            s = Input::question(description + "? [Y/n] ");
        }
        else
        {
            s = Input::question(description + "? [N/y] ");
        }

        if (Str::lower(s) == "y")
        {
            b = true;
        }
        else if (Str::lower(s) == "n")
        {
            b = false;
        }
        else if (s.empty())
        {
            b = defaultValue;
        }

        tmake->cache->addOption(name, b);
    }
}

Compiler *Project::getCompiler()
{
    return compiler;
}

bool Project::getOption(string name)
{
    return tmake->cache->checkOption(name);
}

bool Project::executableExists(string name, bool print)
{
    if (print || tmake->debugMode())
    {
        Output::print("Finding executable: " + name);
    }

    bool found = false;

    vector<string> paths = OS::paths();
    for (unsigned int i = 0; i < paths.size(); i++)
    {
        string path = OS::convertPath(paths[i] + "/" + name);
        if (FS::fileExists(path))
        {
            found = true;
            i = paths.size();
        }
    }

    if (!found && FS::fileExists(name))
    {
        found = true;
    }

    return found;
}

bool Project::packageExists(string name, bool print)
{
    if (print)
    {
        Output::print("Finding package: " + name);
    }

    Package *package = tmake->getPackage(name);
    return package->exists();
}

bool Project::libraryExists(string name, bool print)
{
    if (!tmake->cache->checkLib(name))
    {
        if (print || tmake->debugMode())
        {
            Output::print("Finding library: " + name);
        }

        if (tmake->getProject()->getCompiler()->libraryExists(name))
        {
            tmake->cache->addLib(name);
            return true;
        }
    }
    else
    {
        return true;
    }

    return false;
}

bool Project::includeDirExists(string name, bool print)
{
    if (!tmake->cache->checkIncDir(name))
    {
        if (print || tmake->debugMode())
        {
            Output::print("Finding include dir: " + name);
        }

        if (FS::dirExists(name))
        {
            tmake->cache->addIncDir(name);
            return true;
        }
    }
    else
    {
        return true;
    }

    return false;
}

bool Project::libraryDirExists(string name, bool print)
{
    if (!tmake->cache->checkLibDir(name))
    {
        if (print || tmake->debugMode())
        {
            Output::print("Finding library dir: " + name);
        }

        if (FS::dirExists(name))
        {
            tmake->cache->addLibDir(name);
            return true;
        }
    }
    else
    {
        return true;
    }

    return false;
}

string Project::binDir(string name)
{
    if (tmake->binDir.empty())
    {
        return OS::binDir(tmake->convertName(name));
    }

    return tmake->binDir;
}

string Project::libDir(string name)
{
    if (tmake->libDir.empty())
    {
        return OS::libDir(tmake->convertName(name));
    }

    return tmake->libDir;
}

string Project::incDir(string name)
{
    if (tmake->incDir.empty())
    {
        return OS::incDir(tmake->convertName(name));
    }

    return tmake->incDir;
}

string Project::dataDir(string name)
{
    if (tmake->dataDir.empty())
    {
        return OS::dataDir(tmake->convertName(name));
    }

    return tmake->dataDir;
}

string Project::confDir(string name)
{
    if (tmake->confDir.empty())
    {
        return OS::confDir(tmake->convertName(name));
    }

    return tmake->confDir;
}
