#include <cstdlib>
#include <ctime>
#include <fstream>

#include "headers.hpp"

TMake::TMake()
{
    isDebugMode = false;
    action = "build";

    srcDir = "src"; srcDir.push_back(OS::pathSep());
    headerDir = "include"; headerDir.push_back(OS::pathSep());
    resDir = "res"; resDir.push_back(OS::pathSep());

    globalAction = GlobalActions::Project;
}

TMake::~TMake()
{
}

void TMake::printHelp()
{
    cout << "Usage: tmake [flags] [action]" << endl;
    cout << endl;
    cout << "Flags:" << endl;
    cout << " -h, --help\t\tDisplay this help" << endl;
    cout << " -v, --version\t\tDisplay the version of TMake" << endl;
    cout << " -d, --debug\t\tRun TMake in debug mode" << endl;
    cout << endl;
    cout << " -p, --project\t\tRun the project (this is default)" << endl;
    cout << " -c, --create\t\tCreate a project" << endl;
    cout << endl;
    cout << " --bindir=\t\tSet the binary directory" << endl;
    cout << " --libdir=\t\tSet the library directory" << endl;
    cout << " --incdir=\t\tSet the include directory" << endl;
    cout << " --datadir=\t\tSet the data directory" << endl;
    cout << " --confdir=\t\tSet the conf directory" << endl;
}

bool TMake::parseArg(string arg)
{
    if (arg == "-h" || arg == "--help")
    {
        printHelp();
        exit(0);
    }
    else if (arg == "-v" || arg == "--version")
    {
        cout << "TMake Version " << STR_VERSION << endl;
        exit(0);
    }
    else if (arg == "-d" || arg == "--debug")
    {
        isDebugMode = true;
    }
    else if (arg == "-p" || arg == "--project")
    {
        globalAction = GlobalActions::Project;
    }
    else if (arg == "-c" || arg == "--create")
    {
        globalAction = GlobalActions::Create;
    }
    else if (Str::startsWith(arg, "--bindir="))
    {
        binDir = arg.substr(9, arg.size());
    }
    else if (Str::startsWith(arg, "--libdir="))
    {
        libDir = arg.substr(9, arg.size());
    }
    else if (Str::startsWith(arg, "--incdir="))
    {
        incDir = arg.substr(9, arg.size());
    }
    else if (Str::startsWith(arg, "--confdir="))
    {
        confDir = arg.substr(10, arg.size());
    }
    else if (Str::startsWith(arg, "--datadir="))
    {
        dataDir = arg.substr(10, arg.size());
    }
    else if (arg == "build" || arg == "clean" || arg == "install"
                                              || arg == "uninstall")
    {
        action = arg;
    }
    else
    {
        return false;
    }
    
    return true;
}

bool TMake::init()
{
    Output::debug("Initilising TMake...");
    
    if (!initLua())
    {
        return false;
    }
    
    initLuaFunctions();
    initLuaVariables();

    initCompilers();
    
    OS::init();
    Arch::init();
    
    project = new Project(luaState);
    cache = new Cache();
    
    return true;
}

bool TMake::initLua()
{
    luaState = luaL_newstate();
    if (!luaState)
    {
        return false;
    }
    
    luaL_openlibs(luaState);

#ifdef LUAJIT_SUPPORT
    luaJIT_setmode(luaState, 0, LUAJIT_MODE_ENGINE | LUAJIT_MODE_ON);
#endif
    
    return true;
}

void TMake::initLuaFunctions()
{
    lua_register(luaState, "RequireVersion", Func_RequireVersion);
    
    lua_register(luaState, "Project", Func_Project);

    lua_register(luaState, "SetCompiler", Func_SetCompiler);

    lua_register(luaState, "SrcDir", Func_SrcDir);
    lua_register(luaState, "HeaderDir", Func_HeaderDir);
    lua_register(luaState, "ResDir", Func_ResDir);

    lua_register(luaState, "BinDir", Func_BinDir);
    lua_register(luaState, "LibDir", Func_LibDir);
    lua_register(luaState, "IncDir", Func_IncDir);
    lua_register(luaState, "ConfDir", Func_ConfDir);
    lua_register(luaState, "DataDir", Func_DataDir);

    lua_register(luaState, "CompiledName", Func_CompiledName);

    lua_register(luaState, "GetSources", Func_GetSources);
    lua_register(luaState, "GetHeaders", Func_GetHeaders);
    lua_register(luaState, "GetResources", Func_GetResources);
    lua_register(luaState, "GetLibraries", Func_GetLibraries);
    lua_register(luaState, "GetIncludeDirs", Func_GetIncludeDirs);
    lua_register(luaState, "GetLibraryDirs", Func_GetLibraryDirs);
    lua_register(luaState, "GetDefinitions", Func_GetDefinitions);
    lua_register(luaState, "GetCFlags", Func_GetCFlags);
    lua_register(luaState, "GetLFlags", Func_GetLFlags);
    lua_register(luaState, "GetOption", Func_GetOption);

    lua_register(luaState, "AddSource", Func_AddSource);
    lua_register(luaState, "AddHeader", Func_AddHeader);
    lua_register(luaState, "AddResource", Func_AddResource);
    lua_register(luaState, "AddPackage", Func_AddPackage);
    lua_register(luaState, "AddLibrary", Func_AddLibrary);
    lua_register(luaState, "AddIncludeDir", Func_AddIncludeDir);
    lua_register(luaState, "AddLibraryDir", Func_AddLibraryDir);
    lua_register(luaState, "AddDefinition", Func_AddDefinition);
    lua_register(luaState, "AddCFlag", Func_AddCFlag);
    lua_register(luaState, "AddLFlag", Func_AddLFlag);
    lua_register(luaState, "AddOption", Func_AddOption);

    lua_register(luaState, "CreateExecutable", Func_CreateExecutable);
    lua_register(luaState, "CreateSharedLibrary", Func_CreateSharedLibrary);
    lua_register(luaState, "CreateStaticLibrary", Func_CreateStaticLibrary);

    lua_register(luaState, "FindExecutable", Func_FindExecutable);
    lua_register(luaState, "FindPackage", Func_FindPackage);
    lua_register(luaState, "FindLibrary", Func_FindLibrary);
    lua_register(luaState, "FindIncludeDir", Func_FindIncludeDir);
    lua_register(luaState, "FindLibraryDir", Func_FindLibraryDir);

    lua_register(luaState, "PickExecutable", Func_PickExecutable);
    lua_register(luaState, "PickLibrary", Func_PickLibrary);
    lua_register(luaState, "PickIncludeDir", Func_PickIncludeDir);
    lua_register(luaState, "PickLibraryDir", Func_PickLibraryDir);

    lua_register(luaState, "Execute", Func_Execute);
    lua_register(luaState, "Install", Func_Install);

    lua_register(luaState, "Print", Func_Print);
    lua_register(luaState, "Warning", Func_Warning);

    lua_register(luaState, "RegisterCompiler", Func_RegisterCompiler);
    lua_register(luaState, "AlterCompiler", Func_AlterCompiler);
    lua_register(luaState, "RequiresCompiling", Func_RequiresCompiling);

    lua_register(luaState, "FS_Mkdir", Func_FS_Mkdir);
    lua_register(luaState, "FS_Rm", Func_FS_Rm);
    lua_register(luaState, "FS_Move", Func_FS_Move);
    lua_register(luaState, "FS_Copy", Func_FS_Copy);
    lua_register(luaState, "FS_WriteFile", Func_FS_WriteFile);

    lua_register(luaState, "StrJoin", Func_StrJoin);
    lua_register(luaState, "StrSplit", Func_StrSplit);
    lua_register(luaState, "StrEmpty", Func_StrEmpty);

    lua_register(luaState, "TableHas", Func_TableHas);
}

void TMake::initLuaVariables()
{
    Lua::setStrVar(luaState, "OS", OS::name());
    Lua::setStrVar(luaState, "Arch", Arch::name());
    Lua::setStrVar(luaState, "Action", action);
    Lua::setStrVar(luaState, "BuildDir", Arch::buildDir());
    Lua::setStrVar(luaState, "StrVersion", STR_VERSION);
}

void TMake::initCompilers()
{
    vector<string> dirList = FS::listDir(OS::compilersDir());
    for (unsigned int i = 0; i < dirList.size(); i++)
    {
        string filename = dirList[i];
        if (FS::fileExists(filename))
        {
            Output::debug("Loading compiler: " + filename);

            if (luaL_dofile(luaState, filename.c_str()) != 0)
            {
                Output::warning(Lua::getString(luaState, lua_gettop(luaState)));
            }
        }
    }
}

int TMake::run()
{
    Output::debug("Starting TMake...");
    
    if (!init())
    {
        Output::error("Failed to initilise Lua");
        return 1;
    }

    if (globalAction == GlobalActions::Project)
    {
        if (action == "build")
        {
            Output::status("Building project...");
            FS::mkdir(Arch::buildDir(), false);
        }
        else if (action == "install")
        {
            Output::status("Installing project...");
        }
        else if (action == "uninstall")
        {
            Output::status("Uninstalling project...");
        }
        else if (action == "clean")
        {
            Output::status("Cleaning project...");
        }

        Output::debug("Executing file...");

        if (action == "build")
        {
            int ret = luaL_dofile(luaState, "tmake.lua");
            if (ret != 0)
            {
                Output::error(Lua::getString(luaState, lua_gettop(luaState)));
                end();
                return 1;
            }
        }
        else if (action == "clean")
        {
            for (unsigned int i = 0; i < cache->getGeneratedFiles().size(); i++)
            {
                FS::rm(cache->getGeneratedFiles()[i]);
            }
        }
        else if (action == "install")
        {
            map<string, string> instFiles = cache->getInstallFiles();
            map<string, string>::iterator iter;
            for (iter = instFiles.begin(); iter != instFiles.end(); iter++)
            {
                FS::mkdir(OS::convertPath(iter->second));
                FS::cp(OS::convertPath(iter->first), OS::convertPath(iter->second));
            }
        }
        else if (action == "uninstall")
        {
            map<string, string> instFiles = cache->getInstallFiles();
            map<string, string>::iterator iter;
            for (iter = instFiles.begin(); iter != instFiles.end(); iter++)
            {
                vector<string> files = Str::split(OS::convertPath(iter->first), OS::pathSep());
                files.insert(files.begin(), "");
                for (unsigned int i = 0; i < files.size(); i++)
                {
                    files.erase(files.begin());

                    string strPathSep; strPathSep.push_back(OS::pathSep());

                    string newPath = Str::join(files, strPathSep, "");

                    string fullPath = iter->second + newPath;

                    if (FS::fileExists(fullPath) || FS::dirExists(fullPath))
                    {
                        FS::rm(fullPath);
                    }
                    else if (newPath == "/*")
                    {
                        FS::rm(iter->second);
                    }
                }
            }
        }
    }
    else
    {
        Output::status("Creating project...");

        string name = Input::question("Name: ");
        string computerName = convertName(Str::lower(name));
        Output::print("Computer Name: " + computerName);
        string language = Input::question("Language: ");
        string type = Input::question("Type: ");
        if (type != "exe" && type != "shared" && type != "static")
        {
            type = "exe";
            Output::warning("Executable type chosen");
        }
        string location = Input::question("Location: ");
        if (!FS::dirExists(location))
        {
            Output::error("Location not found, please choose the parent directory to the project");
            end();
            return 1;
        }

        string temp = Input::question("Template: ");
        cout << endl;
        Input::question("Create project? ");

        int ret = luaL_dofile(luaState, OS::convertPath(OS::templatesDir() + temp + ".lua").c_str());
        if (ret != 0)
        {
            Output::error(Lua::getString(luaState, lua_gettop(luaState)));
            end();
            return 1;
        }

        FS::mkdir(OS::convertPath(location));
        FS::mkdir(OS::convertPath(location + "/" + Str::lower(name)));
        FS::mkdir(OS::convertPath(location + "/" + Str::lower(name) + "/src/"));
        FS::mkdir(OS::convertPath(location + "/" + Str::lower(name) + "/include/"));
        FS::mkdir(OS::convertPath(location + "/" + Str::lower(name) + "/res/"));

        location = OS::convertPath(location + "/" + Str::lower(name) + "/");

        lua_getglobal(luaState, (temp + "_Template").c_str());
        Lua::pushString(luaState, name);
        Lua::pushString(luaState, language);
        Lua::pushString(luaState, type);
        Lua::pushString(luaState, location);
        if (lua_pcall(luaState, 4, 10, 0) != 0)
        {
            Output::error(Lua::getString(luaState, lua_gettop(luaState)));
            end();
            return 1;
        }

        vector<string> sources = Lua::getStringList(luaState, -10);
        vector<string> headers = Lua::getStringList(luaState, -9);
        vector<string> resources = Lua::getStringList(luaState, -8);
        vector<string> packages = Lua::getStringList(luaState, -7);
        vector<string> libraries = Lua::getStringList(luaState, -6);
        vector<string> includeDirs = Lua::getStringList(luaState, -5);
        vector<string> libraryDirs = Lua::getStringList(luaState, -4);
        vector<string> definitions = Lua::getStringList(luaState, -3);
        vector<string> cflags = Lua::getStringList(luaState, -2);
        vector<string> lflags = Lua::getStringList(luaState, -1);

        stringstream ss;
        ss << "RequireVersion(1.0)\n\n";
        ss << "Project(\"" << name << "\", \"" << language << "\")\n\n";

        ss << "AddSource(";
        for (unsigned int i = 0; i < sources.size(); i++)
        {
            string src = sources[i];
            ss << "\"" << src << "\"";

            if (i < sources.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddHeader(";
        for (unsigned int i = 0; i < headers.size(); i++)
        {
            string header = headers[i];
            ss << "\"" << header << "\"";

            if (i < headers.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddResource(";
        for (unsigned int i = 0; i < resources.size(); i++)
        {
            string res = resources[i];
            ss << "\"" << res << "\"";

            if (i < resources.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        for (unsigned int i = 0; i < packages.size(); i++)
        {
            string pkg = packages[i];
            ss << "AddPackage(\"" << pkg << "\")\n";
        }

        ss << "AddLibrary(";
        for (unsigned int i = 0; i < libraries.size(); i++)
        {
            string lib = libraries[i];
            ss << "\"" << lib << "\"";

            if (i < libraries.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddIncludeDir(";
        for (unsigned int i = 0; i < includeDirs.size(); i++)
        {
            string dir = includeDirs[i];
            ss << "\"" << dir << "\"";

            if (i < includeDirs.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddLibraryDir(";
        for (unsigned int i = 0; i < libraryDirs.size(); i++)
        {
            string dir = libraryDirs[i];
            ss << "\"" << dir << "\"";

            if (i < libraryDirs.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddDefinition(";
        for (unsigned int i = 0; i < definitions.size(); i++)
        {
            string def = definitions[i];
            ss << "\"" << def << "\"";

            if (i < definitions.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddCFlag(";
        for (unsigned int i = 0; i < cflags.size(); i++)
        {
            string flag = cflags[i];
            ss << "\"" << flag << "\"";

            if (i < cflags.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "AddLFlag(";
        for (unsigned int i = 0; i < lflags.size(); i++)
        {
            string flag = lflags[i];
            ss << "\"" << flag << "\"";

            if (i < lflags.size() - 1)
            {
                ss << ", ";
            }
        }
        ss << ")\n";

        ss << "\n";
        if (type == "exe")
        {
            ss << "CreateExecutable(\"" << Str::lower(name) << "\")\n";
        }
        else if (type == "shared")
        {
            ss << "CreateSharedLibrary(\"" << Str::lower(name) << "\")\n";
        }
        else if (type == "static")
        {
            ss << "CreateStaticLibrary(\"" << Str::lower(name) << "\")\n";
        }

        FS::writeFile(location + "tmake.lua", ss.str());
    }
    
    end();
    
    Output::status("Completed successfully!");
    
    return 0;
}

void TMake::end()
{
    Output::debug("Ending TMake...");
    
    OS::end();
    Arch::end();
    
    if (action == "build")
    {
        cache->setGeneratedFiles(project->generatedFiles);
        cache->setInstallFiles(project->installFiles);
    }

    delete project;
    delete cache;
    
    lua_close(luaState);
}

void TMake::registerCompiler(string name, string language)
{
    Output::debug("Registering " + language + " compiler: " + name);

    Compiler *compiler = new Compiler(luaState, name, language);
    compilers.push_back(compiler);
}

Compiler *TMake::chooseCompiler(string language)
{
    for (unsigned int i = 0; i < compilers.size(); i++)
    {
        Compiler *c = compilers[i];
        if (c->getLanguage() == language)
        {
            return c;
        }
    }

    Lua::error(luaState, "No " + language + " compiler available");
    return NULL;
}

Compiler *TMake::getCompiler(string name, string language)
{
    for (unsigned int i = 0; i < compilers.size(); i++)
    {
        Compiler *c = compilers[i];
        if (c->getName() == name && c->getLanguage() == language)
        {
            return c;
        }
    }

    Lua::error(luaState, "Compiler, " + name + ", not available");
    return NULL;
}

void TMake::createBuildDir(string sourcefile)
{
    string dirName;

    vector<string> paths = Str::split(OS::convertPath(sourcefile), OS::pathSep());
    for (unsigned int i = 0; i < paths.size() - 1; i++)
    {
        dirName.append(paths[i]);
        if (i < paths.size() - 2)
        {
            dirName.push_back(OS::pathSep());
        }
    }

    FS::mkdir(dirName, false);
}

string TMake::convertName(string name)
{
    string newName;

    for (unsigned int i = 0; i < name.size(); i++)
    {
        char c = name[i];

        if (c == ' ')
        {
            newName.push_back('_');
        }
        else
        {
            newName.push_back(c);
        }
    }

    return newName;
}

bool TMake::debugMode()
{
    return isDebugMode;
}

string TMake::getAction()
{
    return action;
}

Project *TMake::getProject()
{
    return project;
}

Package *TMake::getPackage(string name)
{
    if (packages[name])
    {
        return packages[name];
    }
    else
    {
        packages[name] = new Package(name, luaState);
        return packages[name];
    }
}
